# Copyright (c) 2014-2016, The Regents of the University of California.
# Copyright (c) 2016-2017, Nefeli Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the names of the copyright holders nor the names of their
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import copy
import types


def _callback_factory(self, cmd, arg_type):
    return lambda mod, **kwargs: \
        self.bess.run_module_command(self.name, cmd, arg_type, kwargs)


class Module(object):

    def __init__(self, _do_not_create=False, **kwargs):
        self.name = '<uninitialized>'
        self.mclass = self.__class__.__name__

        assert self.__class__.__name__ != 'Module', \
            "Should not instantiate 'Module' directly"

        if 'name' in kwargs:
            name = kwargs['name']
            del kwargs['name']
        else:
            name = None

        if not _do_not_create:
            # create an object
            ret = self.bess.create_module(self.__class__.__name__, name,
                                          self.choose_arg(None, kwargs))
            self.name = ret.name
        else:
            # bind to a pre-existing object, check if it's real
            assert name is not None, "Module should not be None"
            info = self.bess.get_module_info(name)
            assert self.mclass == info.mclass, "Module %s is not of % type" % (
                name, self.mclass)
            self.name = name

        # add mclass-specific methods
        cls = self.bess.get_mclass_info(self.__class__.__name__)
        assert len(cls.cmds) == len(cls.cmd_args)
        for i, cmd in enumerate(cls.cmds):
            func = _callback_factory(self, cmd, cls.cmd_args[i])
            setattr(self, cmd, types.MethodType(func, self))

        self.ogate = None
        self.igate = None

    def __str__(self):
        return '%s::%s' % (self.name, str(self.__class__.__name__))

    def __mul__(self, ogate):
        if not isinstance(ogate, int):
            assert False, 'Gate ID must be an integer'

        if self.ogate is not None:
            assert False, 'Output gate is already bound'

        ret = copy.copy(self)
        ret.ogate = ogate
        return ret

    def __rmul__(self, igate):
        if not isinstance(igate, int):
            assert False, 'Gate ID must be an integer'

        if self.igate is not None:
            assert False, 'Input gate is already bound'

        ret = copy.copy(self)
        ret.igate = igate
        return ret

    def __add__(self, next_mod):
        if not isinstance(next_mod, Module):
            assert False, '%s is not a module' % next_mod

        ogate = 0
        igate = 0

        if self.ogate is not None:
            ogate = self.ogate
            self.ogate = None

        if next_mod.igate is not None:
            igate = next_mod.igate
            next_mod.igate = None

        return self.connect(next_mod, ogate, igate)

    def connect(self, next_mod, ogate=0, igate=0):
        if not isinstance(next_mod, Module):
            assert False, '%s is not a module' % next_mod

        self.bess.connect_modules(self.name, next_mod.name, ogate, igate)

        # for a->b->c syntax
        return next_mod

    # Attach the task numbered `module_taskid` (usually modules only have one
    # task, numbered 0) from this module to a TC tree.
    #
    # The behavior differs based on the arguments provided:
    #
    # * If `wid` is specified, the task is attached as a root in the worker
    #   `wid`.  If `wid` has multiple roots they will be under a default
    #   round-robin policy.
    # * If `parent` is specified, the task is attached as a child of `parent`.
    #   If `parent` is a priority or weighted_fair TC, `priority` or `share`
    #   can be used to customize the child parameter.
    #
    def attach_task(self, parent='', wid=-1, module_taskid=0,
                    priority=None, share=None):
        return self.bess.attach_task(self.name, parent, wid, module_taskid,
                                     priority, share)



# Copyright (c) 2014-2017, The Regents of the University of California.
# Copyright (c) 2016-2017, Nefeli Networks, Inc.
# Copyright (c) 2017, Cloudigo.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the names of the copyright holders nor the names of their
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function
from __future__ import absolute_import

import errno
import grpc
import os
import pprint
import re
import sys
import time

from . import protobuf_to_dict as pb_conv

# pseudo-module multi-importer, used to build module message types
from . import pm_import as _pm

# Ugh: builtin_pb and plugin_pb must be on path, as protoc generates python code
# that assumes it can import files in that directory.
old_path = list(sys.path)
p = os.path.abspath(os.path.dirname(__file__))
if p not in sys.path:
    sys.path.append(p)
for extra in ('builtin_pb', 'plugin_pb'):
    p = os.path.abspath(os.path.join(__file__, '..', extra))
    if p not in sys.path:
        sys.path.append(p)
del extra, p

from builtin_pb import service_pb2_grpc
from builtin_pb import bess_msg_pb2 as bess_msg
from builtin_pb import module_msg_pb2 as module_msg


def _import_modules(name, subdir):
    """Return a module instance retaining just the right protobuf
    names, built by importing most of the *_msg_pb2.py files in the
    builtin_pb and plugin_pb directories, or a subdirectory of them.
    We skip bess_msg_pb2 for historical reasons.

    We detect any name collisions, e.g., if foo_msg_pb2.py defines
    QuuxArg and bar_msg_pb2.py also defines QuuxArg, we catch that
    error here."""
    def protobufs(subdir):
        "Yield modules (subdir=None) or port drivers (subdir='ports')."
        for path in sys.path:
            if not os.path.exists(os.path.join(path, 'builtin_pb')) or \
               not os.path.exists(os.path.join(path, 'plugin_pb')):
                continue
            for directory in ('builtin_pb', 'plugin_pb'):
                if subdir:
                    dirpath = os.path.join(path, directory, subdir)
                    prefix = '{}.{}'.format(directory, subdir)
                else:
                    dirpath = os.path.join(path, directory)
                    prefix = '{}'.format(directory)
                for filename in os.listdir(dirpath):
                    if not filename.endswith('_msg_pb2.py'):
                        continue
                    import_name = '{}.{}'.format(prefix, filename[:-3])
                    if import_name != 'builtin_pb.bess_msg_pb2':
                        yield import_name

    def keep_name(name):
        """
        This defines the set of protobuf names to keep.  Those ending
        with Arg, Response, or Config are kept: Arg for arguments to
        a command, Response for a reply to a command, and Config for
        get_config / set_config state.
        """
        return re.search(r'(?:Arg|Config|Response)$', name) is not None

    try:
        mod = _pm.pm_import(name, protobufs(subdir),
                            name_filter=keep_name,
                            package=__name__)
    except _pm.Collisions as err:
        print('internal error:')
        for key in err.collisions:
            print('', key, 'is defined in', ' and '.join(err.collisions[key]))
        raise SystemExit(1)
    return mod


module_pb = _import_modules('module_pb', None)
port_msg = _import_modules('port_msg', 'ports')
sys.path = old_path
del old_path


def _constraints_to_list(constraint):
    current = 0
    active = []
    while constraint > 0:
        if constraint & 1 == 1:
            active.append(current)
        current += 1
        constraint = constraint >> 1
    return active


class BESS(object):

    class Error(Exception):  # errors from BESS daemon

        def __init__(self, code, errmsg, **kwargs):
            self.code = code
            self.errmsg = errmsg

            # a string->value dict for additional error contexts
            self.info = kwargs

        def __str__(self):
            if self.code in errno.errorcode:
                err_code = errno.errorcode[self.code]
            else:
                err_code = '<unknown>'
            return 'errno=%d (%s: %s), %s' % (
                self.code, err_code, os.strerror(self.code), self.errmsg)

    # abnormal RPC failure
    class RPCError(Exception):
        pass

    # errors from this class itself or misuse of this class
    class APIError(Exception):
        pass

    # An error due to an unsatisfied constraint
    class ConstraintError(Exception):
        pass

    DEF_PORT = 10514
    DEF_GRPC_URL = "localhost:" + str(DEF_PORT)
    BROKEN_CHANNEL = "AbnormalDisconnection"
    CLOSING_CHANNEL = "ConnectionClosing"

    def __init__(self):
        self.debug = False
        self.stub = None
        self.channel = None
        self.peer = self.DEF_GRPC_URL

        self.status = None

    def is_connected(self):
        return (self.status == grpc.ChannelConnectivity.READY)

    def is_connection_broken(self):
        return self.status == self.BROKEN_CHANNEL

    def _update_status(self, connectivity):
        if self.debug:
            print("Channel status: {} -> {}".format(self.status, connectivity))

        # do not update status if previous disconnection is not reported yet
        if self.is_connection_broken():
            return

        if self.status == grpc.ChannelConnectivity.READY and \
                connectivity == grpc.ChannelConnectivity.TRANSIENT_FAILURE:
            self.status = self.BROKEN_CHANNEL
        elif self.status == grpc.ChannelConnectivity.READY and \
                connectivity == grpc.ChannelConnectivity.IDLE:
            # HTTP2 GOAWAY causes premature disconnection. Try to reconnect.
            # This sometimes happens when BESS daemon is launching.
            self.status = self.CLOSING_CHANNEL
        else:
            self.status = connectivity

    def connect(self, grpc_url=DEF_GRPC_URL):
        if self.debug:
            print('Connecting to ' + grpc_url)

        if self.is_connected():
            raise self.APIError('Already connected')

        while not self.is_connected():
            if self.channel is None:
                self.status = None
                self.peer = grpc_url
                self.channel = grpc.insecure_channel(grpc_url)
                self.channel.subscribe(self._update_status,
                                       try_to_connect=True)
                self.stub = service_pb2_grpc.BESSControlStub(self.channel)

            elif self.status == self.CLOSING_CHANNEL:
                self.disconnect()

            elif self.status in [grpc.ChannelConnectivity.TRANSIENT_FAILURE,
                                 grpc.ChannelConnectivity.SHUTDOWN,
                                 self.BROKEN_CHANNEL]:
                self.disconnect()
                raise self.RPCError(
                    'Connection to {} failed'.format(grpc_url))
            time.sleep(0.1)

    # returns no error if already disconnected
    def disconnect(self):
        try:
            if self.channel is not None:
                if self.debug:
                    print('Disconnecting')
                self.channel.unsubscribe(self._update_status)
        finally:
            self.status = None
            self.stub = None
            self.channel = None

    def set_debug(self, flag):
        self.debug = flag

    def _request(self, name, req_pb=None):
        if not self.is_connected():
            if self.is_connection_broken():
                # The channel got abnormally (and asynchronously) disconnected,
                # but RPCError has not been raised yet?
                raise self.RPCError('Broken RPC channel')
            else:
                raise self.APIError('BESS daemon not connected')

        req_fn = getattr(self.stub, name)
        if req_pb is None:
            req_pb = bess_msg.EmptyRequest()

        req_dict = pb_conv.protobuf_to_dict(req_pb)

        if self.debug:
            print('====',  req_fn._method)
            print('--->', type(req_pb).__name__)
            if req_dict:
                pprint.pprint(req_dict)

        try:
            response = req_fn(req_pb)
        except grpc._channel._Rendezvous as e:
            raise self.RPCError(str(e))

        if self.debug:
            print('<---', type(response).__name__)
            res = pb_conv.protobuf_to_dict(response)
            if res:
                pprint.pprint(res)

        if response.error.code != 0:
            code = response.error.code
            errmsg = response.error.errmsg or '(error message is not given)'
            raise self.Error(code, errmsg, query=name, query_arg=req_dict)

        return response

    def kill(self, block=True):
        try:
            response = self._request('KillBess')
        except grpc._channel._Rendezvous:
            pass

        if block:
            while self.is_connected():
                time.sleep(0.1)

        self.disconnect()
        return response

    def get_version(self):
        return self._request('GetVersion')

    def reset_all(self):
        return self._request('ResetAll')

    def pause_all(self):
        return self._request('PauseAll')

    def pause_worker(self, wid):
        request = bess_msg.PauseWorkerRequest()
        request.wid = wid
        return self._request('PauseWorker', request)

    def check_constraints(self):
        response = self.check_scheduling_constraints()
        error = False
        if len(response.violations) != 0 or len(response.modules) != 0:
            print('Placement violations found')
            for violation in response.violations:
                if violation.constraint != 0:
                    valid = ' '.join(
                        map(str, _constraints_to_list(violation.constraint)))
                    print('name %s allowed_sockets [%s] worker_socket %d '
                          'worker_core %d' % (violation.name,
                                              valid,
                                              violation.assigned_node,
                                              violation.assigned_core))
                else:
                    print('name %s has no valid '
                          'placements worker_socket %d '
                          'worker_core %d' % (violation.name,
                                              violation.assigned_node,
                                              violation.assigned_core))
            for module in response.modules:
                print('constraints violated for module %s --'
                      ' please check bessd log' % module.name)
            error = True
        if response.fatal:
            raise self.ConstraintError("Fatal violation of "
                                       "scheduling constraints")
        return error

    def uncheck_resume_all(self):
        return self._request('ResumeAll')

    def check_resume_all(self):
        ret = self.check_constraints()
        self.uncheck_resume_all()
        return ret

    def resume_all(self, check=True):
        if check:
            return self.check_resume_all()
        else:
            return self.uncheck_resume_all()

    def resume_worker(self, wid):
        request = bess_msg.ResumeWorkerRequest()
        request.wid = wid
        return self._request('ResumeWorker', request)

    def check_scheduling_constraints(self):
        return self._request('CheckSchedulingConstraints')

    def list_drivers(self):
        return self._request('ListDrivers')

    def get_driver_info(self, name):
        request = bess_msg.GetDriverInfoRequest()
        request.driver_name = name
        return self._request('GetDriverInfo', request)

    def reset_ports(self):
        return self._request('ResetPorts')

    def list_ports(self):
        return self._request('ListPorts')

    def create_port(self, driver, name=None, arg=None):
        arg = arg or {}

        request = bess_msg.CreatePortRequest()
        request.name = name or ''
        request.driver = driver
        request.num_inc_q = arg.pop('num_inc_q', 0)
        request.num_out_q = arg.pop('num_out_q', 0)
        request.size_inc_q = arg.pop('size_inc_q', 0)
        request.size_out_q = arg.pop('size_out_q', 0)

        message_type = getattr(port_msg, driver + 'Arg', module_msg.EmptyArg)
        arg_msg = pb_conv.dict_to_protobuf(message_type, arg)
        request.arg.Pack(arg_msg)

        return self._request('CreatePort', request)

    def destroy_port(self, name):
        request = bess_msg.DestroyPortRequest()
        request.name = name
        return self._request('DestroyPort', request)

    def set_port_config(self, name, arg):
        request = bess_msg.SetPortConfRequest()
        request.name = name
        # no update is done if mac_addr or mtu is not specified
        request.conf.mac_addr = arg.pop('mac_addr', '00:00:00:00:00:00')
        request.conf.mtu = arg.pop('mtu', 0)
        request.conf.admin_up = arg.pop('admin_up', True)
        return self._request('SetPortConf', request)

    def get_port_config(self, name):
        request = bess_msg.GetPortConfRequest()
        request.name = name
        return self._request('GetPortConf', request)

    def get_port_stats(self, name):
        request = bess_msg.GetPortStatsRequest()
        request.name = name
        return self._request('GetPortStats', request)

    def get_link_status(self, name):
        request = bess_msg.GetLinkStatusRequest()
        request.name = name
        return self._request('GetLinkStatus', request)

    def import_plugin(self, path):
        request = bess_msg.ImportPluginRequest()
        request.path = path
        return self._request('ImportPlugin', request)

    def unload_plugin(self, path):
        request = bess_msg.UnloadPluginRequest()
        request.path = path
        return self._request('UnloadPlugin', request)

    def list_plugins(self):
        return self._request('ListPlugins')

    def list_mclasses(self):
        return self._request('ListMclass')

    def list_modules(self):
        return self._request('ListModules')

    def list_gatehook_classes(self):
        return self._request('ListGateHookClass')

    def list_gatehooks(self):
        return self._request('ListGateHooks')

    def get_gatehook_class_info(self, name):
        request = bess_msg.GetGateHookClassInfoRequest()
        request.name = name
        return self._request('GetGateHookClassInfo', request)

    def get_mclass_info(self, name):
        request = bess_msg.GetMclassInfoRequest()
        request.name = name
        return self._request('GetMclassInfo', request)

    def reset_modules(self):
        return self._request('ResetModules')

    def create_module(self, mclass, name=None, arg=None):
        arg = arg or {}

        request = bess_msg.CreateModuleRequest()
        request.name = name or ''
        request.mclass = mclass

        message_type = getattr(module_pb, mclass + 'Arg', module_msg.EmptyArg)
        arg_msg = pb_conv.dict_to_protobuf(message_type, arg)
        request.arg.Pack(arg_msg)

        return self._request('CreateModule', request)

    def destroy_module(self, name):
        request = bess_msg.DestroyModuleRequest()
        request.name = name
        return self._request('DestroyModule', request)

    def get_module_info(self, name):
        request = bess_msg.GetModuleInfoRequest()
        request.name = name
        return self._request('GetModuleInfo', request)

    def connect_modules(self, m1, m2, ogate=0, igate=0):
        request = bess_msg.ConnectModulesRequest()
        request.m1 = m1
        request.m2 = m2
        request.ogate = ogate
        request.igate = igate
        return self._request('ConnectModules', request)

    def disconnect_modules(self, name, ogate=0):
        request = bess_msg.DisconnectModulesRequest()
        request.name = name
        request.ogate = ogate
        return self._request('DisconnectModules', request)

    def run_module_command(self, name, cmd, arg_type, arg):
        request = bess_msg.CommandRequest()
        request.name = name
        request.cmd = cmd

        try:
            message_type = getattr(module_pb, arg_type)
        except AttributeError as e:
            raise self.APIError('Unknown arg "%s"' % arg_type)

        try:
            arg_msg = pb_conv.dict_to_protobuf(message_type, arg)
        except (KeyError, ValueError) as e:
            raise self.APIError(e)

        request.arg.Pack(arg_msg)

        try:
            response = self._request('ModuleCommand', request)
        except self.Error as e:
            e.info.update(module=name, command=cmd, command_arg=arg)
            raise

        if response.HasField('data'):
            response_type_str = response.data.type_url.split('.')[-1]
            response_type = getattr(module_pb, response_type_str,
                                    module_msg.EmptyArg)
            result = response_type()
            response.data.Unpack(result)
            return result
        else:
            return response

    # It might be nice if we could name hook instances directly,
    # rather than using <hook, module, direction, gate> tuples...
    def run_gatehook_command(self, name, mod, direction, gate, cmd,
                             arg_type, arg):
        request = bess_msg.GateHookCommandRequest()
        request.hook.hook_name = name
        request.hook.module_name = mod
        if direction == 'in':
            request.hook.igate = gate
        elif direction == 'out' or direction is None:
            request.hook.ogate = gate
        else:
            raise self.APIError('direction must be either "out" or "in"')
        request.cmd = cmd

        try:
            message_type = getattr(module_pb, arg_type)
        except AttributeError as e:
            raise self.APIError('Unknown arg "%s"' % arg_type)

        try:
            arg_msg = pb_conv.dict_to_protobuf(message_type, arg)
        except (KeyError, ValueError) as e:
            raise self.APIError(e)

        request.hook.arg.Pack(arg_msg)

        try:
            response = self._request('GateHookCommand', request)
        except self.Error as e:
            e.info.update(hook_name=name, module_name=mod, direction=direction,
                          gate=gate, command=cmd, command_arg=arg)
            raise

        if response.HasField('data'):
            response_type_str = response.data.type_url.split('.')[-1]
            response_type = getattr(module_pb, response_type_str,
                                    module_msg.EmptyArg)
            result = response_type()
            response.data.Unpack(result)
            return result
        else:
            return response

    def _configure_gate_hook(self, hook_class, hook_name, module, arg,
                             enable=None, direction=None, gate=None):
        if gate is None:
            gate = -1
        if direction is None:
            direction = 'out'
        if enable is None:
            enable = False
        request = bess_msg.ConfigureGateHookRequest()
        request.hook.class_name = hook_class
        request.hook.hook_name = hook_name
        request.hook.module_name = module
        request.enable = enable
        if direction == 'in':
            request.hook.igate = gate
        elif direction == 'out':
            request.hook.ogate = gate
        else:
            raise self.APIError('direction must be either "out" or "in"')
        request.hook.arg.Pack(arg)
        return self._request('ConfigureGateHook', request)

    def configure_resume_hook(self, name, arg, enable=True):
        if enable is None:
            enable = True
        request = bess_msg.ConfigureResumeHookRequest()
        request.hook_name = name
        request.enable = enable
        request.arg.Pack(arg)
        return self._request('ConfigureResumeHook', request)

    def tcpdump_gate(self, enable, name, m, direction='out', gate=0, fifo=None):
        arg = bess_msg.TcpdumpArg()
        if fifo is not None:
            arg.fifo = fifo
        return self._configure_gate_hook('TcpDump', name, m, arg, enable,
                                         direction, gate)

    def track_gate(self, enable, name, m, bits=False, direction='out',
                   gate=-1):
        arg = bess_msg.TrackArg()
        arg.bits = bits
        return self._configure_gate_hook('Track', name, m, arg, enable,
                                         direction, gate)

    def pcapng_gate(self, enable, name, m, direction='out', gate=0, fifo=None):
        arg = bess_msg.PcapngArg()
        if fifo is not None:
            arg.fifo = fifo
        return self._configure_gate_hook('PcapNg', name, m, arg, enable,
                                         direction, gate)

    def list_workers(self):
        return self._request('ListWorkers')

    def add_worker(self, wid, core, scheduler=None):
        request = bess_msg.AddWorkerRequest()
        request.wid = wid
        request.core = core
        request.scheduler = scheduler or ''
        return self._request('AddWorker', request)

    def destroy_worker(self, wid):
        request = bess_msg.DestroyWorkerRequest()
        request.wid = wid
        return self._request('DestroyWorker', request)

    def list_tcs(self, wid=-1):
        request = bess_msg.ListTcsRequest()
        request.wid = wid

        return self._request('ListTcs', request)

    def add_tc(self, name, policy, wid=-1, parent='', resource=None,
               priority=None, share=None, limit=None, max_burst=None,
               leaf_module_name=None, leaf_module_taskid=None):
        request = bess_msg.AddTcRequest()
        class_ = getattr(request, 'class')
        class_.parent = parent
        class_.name = name
        class_.wid = wid
        class_.policy = policy

        if priority is not None:
            class_.priority = priority

        if share is not None:
            class_.share = share

        if resource is not None:
            class_.resource = resource

        if limit:
            for k in limit:
                class_.limit[k] = limit[k]

        if max_burst:
            for k in max_burst:
                class_.max_burst[k] = max_burst[k]
        if leaf_module_name is not None:
            class_.leaf_module_name = leaf_module_name
        if leaf_module_taskid is not None:
            class_.leaf_module_taskid = leaf_module_taskid

        return self._request('AddTc', request)

    def update_tc_params(self, name, resource=None, limit=None, max_burst=None,
                         leaf_module_name=None, leaf_module_taskid=0):
        request = bess_msg.UpdateTcParamsRequest()
        class_ = getattr(request, 'class')
        class_.name = name
        if resource is not None:
            class_.resource = resource

        if limit:
            for k in limit:
                class_.limit[k] = limit[k]

        if max_burst:
            for k in max_burst:
                class_.max_burst[k] = max_burst[k]

        if leaf_module_name is not None:
            class_.leaf_module_name = leaf_module_name
        if leaf_module_taskid is not None:
            class_.leaf_module_taskid = leaf_module_taskid

        return self._request('UpdateTcParams', request)

    # Attach the task numbered `module_taskid` (usually modules only have one
    # task, numbered 0) from the module `module_name` to a TC tree.
    #
    # The behavior differs based on the arguments provided:
    #
    # * If `wid` is specified, the task is attached as a root in the worker
    #   `wid`.  If `wid` has multiple roots they will be under a default
    #   round-robin policy.
    # * If `parent` is specified, the task is attached as a child of `parent`.
    #   If `parent` is a priority or weighted_fair TC, `priority` or `share`
    #   can be used to customize the child parameter.
    #
    def attach_task(self, module_name, parent='', wid=-1,
                    module_taskid=0, priority=None, share=None):
        request = bess_msg.UpdateTcParentRequest()
        class_ = getattr(request, 'class')
        class_.leaf_module_name = module_name
        class_.leaf_module_taskid = module_taskid
        class_.parent = parent
        class_.wid = wid

        if priority is not None:
            class_.priority = priority

        if share is not None:
            class_.share = share

        return self._request('UpdateTcParent', request)

    # Deprecated alias for attach_task
    def attach_module(self, *args, **kwargs):
        return self.attach_task(*args, **kwargs)

    def get_tc_stats(self, name):
        request = bess_msg.GetTcStatsRequest()
        request.name = name
        return self._request('GetTcStats', request)

    def dump_mempool(self, socket=-1):
        request = bess_msg.DumpMempoolRequest()
        request.socket = socket
        return self._request('DumpMempool', request)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ports/port_msg.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ports/port_msg.proto',
  package='bess.pb',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x14ports/port_msg.proto\x12\x07\x62\x65ss.pb\"\x1a\n\x0bPCAPPortArg\x12\x0b\n\x03\x64\x65v\x18\x01 \x01(\t\"\xb5\x01\n\nPMDPortArg\x12\x10\n\x08loopback\x18\x01 \x01(\x08\x12\x11\n\x07port_id\x18\x02 \x01(\x04H\x00\x12\r\n\x03pci\x18\x03 \x01(\tH\x00\x12\x0e\n\x04vdev\x18\x04 \x01(\tH\x00\x12\x1d\n\x15vlan_offload_rx_strip\x18\x05 \x01(\x08\x12\x1e\n\x16vlan_offload_rx_filter\x18\x06 \x01(\x08\x12\x1c\n\x14vlan_offload_rx_qinq\x18\x07 \x01(\x08\x42\x06\n\x04port\"V\n\x11UnixSocketPortArg\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x1a\n\x12min_rx_interval_ns\x18\x02 \x01(\x03\x12\x17\n\x0f\x63onfirm_connect\x18\x03 \x01(\x08\"\xba\x01\n\x08VPortArg\x12\x0e\n\x06ifname\x18\x01 \x01(\t\x12\x10\n\x06\x64ocker\x18\x02 \x01(\tH\x00\x12\x17\n\rcontainer_pid\x18\x03 \x01(\x03H\x00\x12\x0f\n\x05netns\x18\x04 \x01(\tH\x00\x12\x10\n\x08rxq_cpus\x18\x05 \x03(\x03\x12\x0e\n\x06tx_tci\x18\x06 \x01(\x04\x12\x14\n\x0ctx_outer_tci\x18\x07 \x01(\x04\x12\x10\n\x08loopback\x18\x08 \x01(\x08\x12\x10\n\x08ip_addrs\x18\t \x03(\tB\x06\n\x04\x63pidb\x06proto3'
)




_PCAPPORTARG = _descriptor.Descriptor(
  name='PCAPPortArg',
  full_name='bess.pb.PCAPPortArg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dev', full_name='bess.pb.PCAPPortArg.dev', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33,
  serialized_end=59,
)


_PMDPORTARG = _descriptor.Descriptor(
  name='PMDPortArg',
  full_name='bess.pb.PMDPortArg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='loopback', full_name='bess.pb.PMDPortArg.loopback', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='port_id', full_name='bess.pb.PMDPortArg.port_id', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='pci', full_name='bess.pb.PMDPortArg.pci', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vdev', full_name='bess.pb.PMDPortArg.vdev', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vlan_offload_rx_strip', full_name='bess.pb.PMDPortArg.vlan_offload_rx_strip', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vlan_offload_rx_filter', full_name='bess.pb.PMDPortArg.vlan_offload_rx_filter', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vlan_offload_rx_qinq', full_name='bess.pb.PMDPortArg.vlan_offload_rx_qinq', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='port', full_name='bess.pb.PMDPortArg.port',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=62,
  serialized_end=243,
)


_UNIXSOCKETPORTARG = _descriptor.Descriptor(
  name='UnixSocketPortArg',
  full_name='bess.pb.UnixSocketPortArg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='bess.pb.UnixSocketPortArg.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='min_rx_interval_ns', full_name='bess.pb.UnixSocketPortArg.min_rx_interval_ns', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='confirm_connect', full_name='bess.pb.UnixSocketPortArg.confirm_connect', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=245,
  serialized_end=331,
)


_VPORTARG = _descriptor.Descriptor(
  name='VPortArg',
  full_name='bess.pb.VPortArg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ifname', full_name='bess.pb.VPortArg.ifname', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='docker', full_name='bess.pb.VPortArg.docker', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='container_pid', full_name='bess.pb.VPortArg.container_pid', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='netns', full_name='bess.pb.VPortArg.netns', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rxq_cpus', full_name='bess.pb.VPortArg.rxq_cpus', index=4,
      number=5, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx_tci', full_name='bess.pb.VPortArg.tx_tci', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tx_outer_tci', full_name='bess.pb.VPortArg.tx_outer_tci', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='loopback', full_name='bess.pb.VPortArg.loopback', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ip_addrs', full_name='bess.pb.VPortArg.ip_addrs', index=8,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='cpid', full_name='bess.pb.VPortArg.cpid',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=334,
  serialized_end=520,
)

_PMDPORTARG.oneofs_by_name['port'].fields.append(
  _PMDPORTARG.fields_by_name['port_id'])
_PMDPORTARG.fields_by_name['port_id'].containing_oneof = _PMDPORTARG.oneofs_by_name['port']
_PMDPORTARG.oneofs_by_name['port'].fields.append(
  _PMDPORTARG.fields_by_name['pci'])
_PMDPORTARG.fields_by_name['pci'].containing_oneof = _PMDPORTARG.oneofs_by_name['port']
_PMDPORTARG.oneofs_by_name['port'].fields.append(
  _PMDPORTARG.fields_by_name['vdev'])
_PMDPORTARG.fields_by_name['vdev'].containing_oneof = _PMDPORTARG.oneofs_by_name['port']
_VPORTARG.oneofs_by_name['cpid'].fields.append(
  _VPORTARG.fields_by_name['docker'])
_VPORTARG.fields_by_name['docker'].containing_oneof = _VPORTARG.oneofs_by_name['cpid']
_VPORTARG.oneofs_by_name['cpid'].fields.append(
  _VPORTARG.fields_by_name['container_pid'])
_VPORTARG.fields_by_name['container_pid'].containing_oneof = _VPORTARG.oneofs_by_name['cpid']
_VPORTARG.oneofs_by_name['cpid'].fields.append(
  _VPORTARG.fields_by_name['netns'])
_VPORTARG.fields_by_name['netns'].containing_oneof = _VPORTARG.oneofs_by_name['cpid']
DESCRIPTOR.message_types_by_name['PCAPPortArg'] = _PCAPPORTARG
DESCRIPTOR.message_types_by_name['PMDPortArg'] = _PMDPORTARG
DESCRIPTOR.message_types_by_name['UnixSocketPortArg'] = _UNIXSOCKETPORTARG
DESCRIPTOR.message_types_by_name['VPortArg'] = _VPORTARG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PCAPPortArg = _reflection.GeneratedProtocolMessageType('PCAPPortArg', (_message.Message,), {
  'DESCRIPTOR' : _PCAPPORTARG,
  '__module__' : 'ports.port_msg_pb2'
  # @@protoc_insertion_point(class_scope:bess.pb.PCAPPortArg)
  })
_sym_db.RegisterMessage(PCAPPortArg)

PMDPortArg = _reflection.GeneratedProtocolMessageType('PMDPortArg', (_message.Message,), {
  'DESCRIPTOR' : _PMDPORTARG,
  '__module__' : 'ports.port_msg_pb2'
  # @@protoc_insertion_point(class_scope:bess.pb.PMDPortArg)
  })
_sym_db.RegisterMessage(PMDPortArg)

UnixSocketPortArg = _reflection.GeneratedProtocolMessageType('UnixSocketPortArg', (_message.Message,), {
  'DESCRIPTOR' : _UNIXSOCKETPORTARG,
  '__module__' : 'ports.port_msg_pb2'
  # @@protoc_insertion_point(class_scope:bess.pb.UnixSocketPortArg)
  })
_sym_db.RegisterMessage(UnixSocketPortArg)

VPortArg = _reflection.GeneratedProtocolMessageType('VPortArg', (_message.Message,), {
  'DESCRIPTOR' : _VPORTARG,
  '__module__' : 'ports.port_msg_pb2'
  # @@protoc_insertion_point(class_scope:bess.pb.VPortArg)
  })
_sym_db.RegisterMessage(VPortArg)


# @@protoc_insertion_point(module_scope)



# Copyright (c) 2014-2016, The Regents of the University of California.
# Copyright (c) 2016-2017, Nefeli Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the names of the copyright holders nor the names of their
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

class Port(object):

    def __init__(self, **kwargs):
        self.name = '<uninitialized>'
        self.driver = self.__class__.__name__

        assert self.driver != 'Port', "Do not instantiate 'Port' directly"

        name = kwargs.pop('name', None)

        ret = self.bess.create_port(self.driver, name,
                                    self.choose_arg(None, kwargs))

        self.name = ret.name
        self.mac_addr = ret.mac_addr

    def __str__(self):
        return '%s/%s' % (self.name, self.driver)

    def get_port_stats(self):
        return self.bess.get_port_stats(self.name)

    def get_link_status(self):
        return self.bess.get_link_status(self.name)

    def set_port_config(self, **kwargs):
        return self.bess.set_port_config(self.name,
                                         self.choose_arg(None, kwargs))

    def get_port_config(self):
        return self.bess.get_port_config(self.name)



# Copyright (c) 2014-2016, The Regents of the University of California.
# Copyright (c) 2016-2017, Nefeli Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the names of the copyright holders nor the names of their
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import
import os
import sys
import unittest

# See bess.py for why this is here.
bipath = os.path.abspath(os.path.join(__file__, '..', 'builtin_pb'))
if bipath not in sys.path:
    sys.path.insert(1, bipath)
del bipath

from . import protobuf_to_dict as pb_conv
from .builtin_pb import bess_msg_pb2 as bess_msg
from .builtin_pb import test_msg_pb2 as test_msg


class TestProtobufConvert(unittest.TestCase):

    def test_protobuf_to_dict(self):
        pb = bess_msg.CreatePortResponse()
        pb.error.code = 1
        pb.error.errmsg = 'bar'
        pb.name = 'foo'

        true_result = {
            'error': {
                'code': 1,
                'errmsg': 'bar',
            },
            'name': 'foo'
        }
        result = pb_conv.protobuf_to_dict(pb)
        self.assertEqual(true_result, result)

    def test_dict_to_protobuf(self):
        pb = bess_msg.CreatePortResponse()
        pb.error.code = 1
        pb.error.errmsg = 'bar'
        pb.name = 'foo'

        msg_dict = {
            'error': {
                'code': 1,
                'errmsg': 'bar',
            },
            'name': 'foo'
        }

        msg = pb_conv.dict_to_protobuf(bess_msg.CreatePortResponse, msg_dict)
        self.assertEqual(msg, pb)

        pb = bess_msg.CreateModuleRequest()
        pb.name = 'm1'
        pb.mclass = 'bpf'

        kv = {
            'name': 'm1',
            'mclass': 'bpf',
        }
        msg = pb_conv.dict_to_protobuf(bess_msg.CreateModuleRequest, kv)
        self.assertEqual(msg, pb)

    def test_dict_to_protobuf_msg_has_nested_dict(self):
        pb = test_msg.NestedDictMsg()

        # pb.a = test_msg.UnnestedDictMsg()
        a_dict = {7: 1, 8: 10}
        init_from_dict(pb.a.dict, a_dict)

        b_dict = {7: 5}
        init_from_dict(pb.b.dict, b_dict)

        c_dict = {110: 10, 7: 5, 6: 7}
        init_from_dict(pb.c.dict, c_dict)

        msg_dict = {
            'a': {'dict': a_dict},
            'b': {'dict': b_dict},
            'c': {'dict': c_dict},
        }

        msg = pb_conv.dict_to_protobuf(test_msg.NestedDictMsg, msg_dict)
        self.assertEqual(msg, pb)

    def test_dict_to_protobuf_msg_is_unnested_dict(self):
        pb = test_msg.UnnestedDictMsg()
        init_from_dict(pb.dict, {0: 10, 1: 5, 6: 7})

        msg_dict = {
            'dict': {0: 10, 1: 5, 6: 7}
        }

        msg = pb_conv.dict_to_protobuf(test_msg.UnnestedDictMsg, msg_dict)
        self.assertEqual(msg, pb)


def init_from_dict(pb_map, d):
    for k, v in d.items():
        pb_map[k] = v



#!/usr/bin/env python
import os
import glob
import sys
import pkg_resources

from setuptools import setup, find_namespace_packages
from distutils.command.clean import clean as _clean
from distutils.command.build_py import build_py as _build_py

PROTOBUF_DIR = 'protobuf'
PB2_OUTPUT_DIR = 'builtin_pb'

def generate_proto(source):
    """Invokes the Protocol Compiler to generate _pb2.py and _pb2_grpc.py
    from the given .proto file.  Does nothing if the output already exists
    and is newer than the input."""

    current_dir = (os.path.dirname(__file__) or os.curdir)

    output = os.path.join(
        current_dir,
        source.replace(".proto", "_pb2.py").replace(PROTOBUF_DIR, PB2_OUTPUT_DIR)
    )

    if (not os.path.exists(output) or
       (os.path.exists(source) and
       os.path.getmtime(source) > os.path.getmtime(output))):
        print("Generating %s..." % output)

        if not os.path.exists(source):
            sys.stderr.write("Can't find required file: %s\n" % source)
            sys.exit(-1)

        from grpc.tools import protoc
        proto_include = pkg_resources.resource_filename('grpc_tools', '_proto')
        command = [
            'grpc_tools.protoc',
            '-I{}'.format(proto_include),
            '--proto_path=protobuf',
            '--python_out={}/{}/'.format(current_dir, PB2_OUTPUT_DIR),
            '--grpc_python_out={}/{}/'.format(current_dir, PB2_OUTPUT_DIR),
        ] + [source]
        if protoc.main(command) != 0:
            sys.exit(-1)

class build_py(_build_py):
  def run(self):
    files = glob.glob(os.path.join(PROTOBUF_DIR, '*.proto'))
    files = files + glob.glob(os.path.join(PROTOBUF_DIR, 'ports', '*.proto'))
    for name in files:
        generate_proto(name)
    _build_py.run(self)

class clean(_clean):
    def run(self):
        for (dirpath, dirnames, filenames) in os.walk("."):
            print(filenames)
            for filename in filenames:
                filepath = os.path.join(dirpath, filename)
                if filepath.endswith("_pb2.py") or filepath.endswith("_pb2_grpc.py"):
                    os.remove(filepath)
        _clean.run(self)

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pybess_grpc',
    version='0.0.4-3',
    description='Python Bindings to Interact with BESS GRPC Daemon',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Remi Vichery',
    author_email='remi.vichery@gmail.com',
    packages=find_namespace_packages(),
    cmdclass={
        'clean': clean,
        'build_py': build_py
    },
    setup_requires=[
        "grpcio-tools==1.27.2",
        "grpcio==1.27.2",
        "protobuf==3.11.3"
    ],
    install_requires=[
        "setuptools",
        "grpcio==1.27.2",
        "grpcio-tools==1.27.2",
        "protobuf==3.11.3",
    ],
)